from asyncio import Lock
from contextlib import AsyncExitStack, asynccontextmanager
from collections.abc import Sequence, AsyncIterator
from pathlib import PurePath
from typing import override

from .exceptions import (
    InvalidServiceError,
    NodeExistsError,
    UnauthorizedError,
)
from ._lib import (
    else_none,
    is_valid_name,
    normalize_path,
)
from .types import (
    ChangeAction,
    CreateFileService,
    CreateFileServiceMiddleware,
    CreateHasher,
    CreateService,
    CreateServiceMiddleware,
    CreateSnapshotService,
    CreateSnapshotServiceMiddleware,
    Drive,
    FileService,
    MediaInfo,
    Node,
    ReadableFile,
    Service,
    SnapshotService,
    WritableFile,
)


_API_VERSION = 4


@asynccontextmanager
async def create_drive(
    *,
    file: CreateFileService,
    snapshot: CreateSnapshotService,
    file_middleware: Sequence[CreateFileServiceMiddleware] | None = None,
    snapshot_middleware: Sequence[CreateSnapshotServiceMiddleware] | None = None,
) -> AsyncIterator[Drive]:
    async with _create_service(
        create_service=snapshot,
        middleware_list=snapshot_middleware,
    ) as snapshot_service, _create_service(
        create_service=file, middleware_list=file_middleware
    ) as file_service:
        yield _DefaultDrive(
            file_service=file_service, snapshot_service=snapshot_service
        )


@asynccontextmanager
async def _create_service[
    T: Service
](
    create_service: CreateService[T],
    middleware_list: Sequence[CreateServiceMiddleware[T]] | None,
):
    async with AsyncExitStack() as stack:
        service = await stack.enter_async_context(create_service())
        if service.api_version != _API_VERSION:
            raise InvalidServiceError(
                f"invalid version: required {_API_VERSION}, got {service.api_version}"
            )

        if not middleware_list:
            middleware_list = []

        for create_middleware in middleware_list:
            service = await stack.enter_async_context(create_middleware(service))
            if service.api_version != _API_VERSION:
                raise InvalidServiceError(
                    f"invalid version: required {_API_VERSION}, got {service.api_version}"
                )

        yield service


class _DefaultDrive(Drive):
    def __init__(
        self,
        *,
        file_service: FileService,
        snapshot_service: SnapshotService,
    ) -> None:
        self._sync_lock = Lock()
        self._fs = file_service
        self._ss = snapshot_service

    @override
    async def get_root(self) -> Node:
        return await self._ss.get_root()

    @override
    async def get_node_by_id(self, node_id: str) -> Node:
        return await self._ss.get_node_by_id(node_id)

    @override
    async def get_node_by_path(self, path: PurePath) -> Node:
        path = normalize_path(path)
        return await self._ss.get_node_by_path(path)

    @override
    async def resolve_path(self, node: Node) -> PurePath:
        return await self._ss.resolve_path_by_id(node.id)

    @override
    async def get_child_by_name(
        self,
        name: str,
        parent: Node,
    ) -> Node:
        return await self._ss.get_child_by_name(name, parent.id)

    @override
    async def get_children(self, parent: Node) -> list[Node]:
        return await self._ss.get_children_by_id(parent.id)

    @override
    async def get_trashed_nodes(self, flatten: bool = False) -> list[Node]:
        rv = await self._ss.get_trashed_nodes()
        if flatten:
            return rv

        ancestor_set = set(_.id for _ in rv if _.is_directory)
        if not ancestor_set:
            return rv

        tmp: list[Node] = []
        for node in rv:
            if not await _in_ancestor_set(self, node, ancestor_set):
                tmp.append(node)
        return tmp

    @override
    async def find_nodes_by_regex(self, pattern: str) -> list[Node]:
        return await self._ss.find_nodes_by_regex(pattern)

    @override
    async def walk(
        self,
        node: Node,
        *,
        include_trashed: bool = False,
    ) -> AsyncIterator[tuple[Node, list[Node], list[Node]]]:
        if not node.is_directory:
            return
        q = [node]
        while q:
            node = q[0]
            del q[0]
            if not include_trashed and node.is_trashed:
                continue

            children = await self.get_children(node)
            directorys: list[Node] = []
            files: list[Node] = []
            for child in children:
                if not include_trashed and child.is_trashed:
                    continue
                if child.is_directory:
                    directorys.append(child)
                else:
                    files.append(child)
            yield node, directorys, files

            q.extend(directorys)

    @override
    async def create_directory(
        self,
        name: str,
        parent: Node,
        *,
        exist_ok: bool = False,
    ) -> Node:
        # sanity check
        if not parent.is_directory:
            raise ValueError("parent is not a directory")
        if not name:
            raise ValueError("directory name is empty")
        if not is_valid_name(name):
            raise ValueError("no `/` or `\\` allowed in directory name")
        if not await self.is_authorized():
            raise UnauthorizedError()

        if not exist_ok:
            node = await else_none(
                self.get_child_by_name(
                    name,
                    parent,
                )
            )
            if node:
                raise NodeExistsError(node)

        return await self._fs.create_directory(
            name,
            parent,
            exist_ok=exist_ok,
            private=None,
        )

    @asynccontextmanager
    @override
    async def download_file(self, node: Node) -> AsyncIterator[ReadableFile]:
        # sanity check
        if node.is_directory:
            raise ValueError("node should be a file")
        if not await self.is_authorized():
            raise UnauthorizedError()

        async with self._fs.download_file(node) as file:
            yield file

    @asynccontextmanager
    @override
    async def upload_file(
        self,
        name: str,
        parent: Node,
        *,
        size: int | None = None,
        mime_type: str | None = None,
        media_info: MediaInfo | None = None,
    ) -> AsyncIterator[WritableFile]:
        # sanity check
        if not parent.is_directory:
            raise ValueError("parent is not a directory")
        if not name:
            raise ValueError("directory name is empty")
        if not is_valid_name(name):
            raise ValueError("no `/` or `\\` allowed in directory name")
        if not await self.is_authorized():
            raise UnauthorizedError()

        node = await else_none(self.get_child_by_name(name, parent))
        if node:
            raise NodeExistsError(node)

        async with self._fs.upload_file(
            name,
            parent,
            size=size,
            mime_type=mime_type,
            media_info=media_info,
            private=None,
        ) as file:
            yield file

    @override
    async def purge_trash(self) -> None:
        # sanity check
        if not await self.is_authorized():
            raise UnauthorizedError()

        await self._fs.purge_trash()

    @override
    async def move(
        self,
        node: Node,
        *,
        new_parent: Node | None = None,
        new_name: str | None = None,
        trashed: bool | None = None,
    ) -> Node:
        # sanity check
        if not await self.is_authorized():
            raise UnauthorizedError()

        if not new_parent and not new_name and trashed is None:
            raise ValueError("nothing to move")

        if new_name and not is_valid_name(new_name):
            raise ValueError("no `/` or `\\` allowed in file name")

        root_node = await self.get_root()
        if node.id == root_node.id:
            raise ValueError("root node is immutable")

        if new_parent:
            if new_parent.is_trashed != node.is_trashed:
                raise ValueError("cannot move accross trash")
            if not new_parent.is_directory:
                raise ValueError("new_parent is not a directory")
            if await _contains(self, node, new_parent):
                raise ValueError("new_parent is a descendant of the source node")

        return await self._fs.move(
            node,
            new_parent=new_parent,
            new_name=new_name,
            trashed=trashed,
        )

    @override
    async def sync(self) -> AsyncIterator[ChangeAction]:
        if not await self.is_authorized():
            raise UnauthorizedError()

        async with self._sync_lock:
            initial_cursor = await self._fs.get_initial_cursor()

            cursor = await self._ss.get_current_cursor()
            if not cursor:
                cursor = initial_cursor

            # no data before, get the root node and cache it
            if cursor == initial_cursor:
                node = await self._fs.get_root()
                await self._ss.set_root(node)

            async for changes, next_ in self._fs.get_changes(cursor):
                await self._ss.apply_changes(changes, next_)

                for change in changes:
                    yield change

    @override
    async def get_hasher_factory(self) -> CreateHasher:
        return await self._fs.get_hasher_factory()

    @override
    async def is_authorized(self) -> bool:
        return await self._fs.is_authorized()

    @override
    async def get_oauth_url(self) -> str:
        return await self._fs.get_oauth_url()

    @override
    async def set_oauth_token(self, token: str):
        return await self._fs.set_oauth_token(token)


async def _in_ancestor_set(drive: Drive, node: Node, ancestor_set: set[str]) -> bool:
    if node.parent_id is None:
        return False
    parent = await drive.get_node_by_id(node.parent_id)
    if not parent:
        return False
    if parent.id in ancestor_set:
        return True
    included = await _in_ancestor_set(drive, parent, ancestor_set)
    if included:
        ancestor_set.add(parent.id)
    return included


async def _contains(drive: Drive, ancestor: Node, node: Node) -> bool:
    visited: set[str] = set()
    while True:
        if ancestor.id == node.id:
            # meet the ancestor
            return True
        if not node.parent_id:
            # reached the root but never meet the ancestor
            return False

        visited.add(node.id)
        node = await drive.get_node_by_id(node.parent_id)
        if node.id in visited:
            raise RuntimeError("detected node cycle")



from logging import getLogger
from pathlib import Path, PurePath
from typing import BinaryIO
import asyncio
import os

from .exceptions import (
    NodeExistsError,
    NodeNotFoundError,
    IsADirectoryError,
)
from ._lib import (
    else_none,
    resolve_path,
)
from .types import (
    Drive,
    MediaInfo,
    Node,
    ReadableFile,
    WritableFile,
)


__all__ = (
    "download_file_to_local",
    "find_duplicate_nodes",
    "move_node",
    "upload_file_from_local",
)


_DEFAULT_FILE_MIME_TYPE = "application/octet-stream"
_CHUNK_SIZE = 256 * 1024


async def move_node(
    drive: Drive,
    src_path: PurePath,
    dst_path: PurePath,
) -> Node:
    src_node = await drive.get_node_by_path(src_path)

    # case 1 - move to a relative path
    if not dst_path.is_absolute():
        # case 1.1 - a name, not path
        if dst_path.name == dst_path:
            # case 1.1.1 - move to the same directory, do nothing
            if dst_path.name == ".":
                return src_node
            # case 1.1.2 - rename only
            if dst_path.name != "..":
                return await drive.move(
                    src_node,
                    new_parent=None,
                    new_name=dst_path.name,
                )
            # case 1.1.3 - move to parent directory, the same as case 1.2

        # case 1.2 - a relative path, resolve to absolute path
        # NOTE PurePath does not implement normalizing algorithm
        dst_path = resolve_path(src_path.parent, dst_path)

    # case 2 - move to an absolute path
    dst_node = await else_none(drive.get_node_by_path(dst_path))
    # case 2.1 - the destination is empty
    if not dst_node:
        # move to the parent directory of the destination
        try:
            new_parent = await drive.get_node_by_path(dst_path.parent)
        except NodeNotFoundError as e:
            raise ValueError(f"no direct path to {dst_path}") from e
        return await drive.move(src_node, new_parent=new_parent, new_name=dst_path.name)
    # case 2.2 - the destination is a file
    if not dst_node.is_directory:
        # do not overwrite existing file
        raise NodeExistsError(dst_node)
    # case 2.3 - the distination is a directory
    return await drive.move(src_node, new_parent=dst_node, new_name=None)


async def find_duplicate_nodes(
    drive: Drive,
    root_node: Node | None = None,
) -> list[list[Node]]:
    if not root_node:
        root_node = await drive.get_root()

    rv: list[list[Node]] = []
    async for _root, directorys, files in drive.walk(root_node):
        nodes = directorys + files
        seen: dict[str, list[Node]] = {}
        for node in nodes:
            if node.name not in seen:
                seen[node.name] = [node]
            else:
                seen[node.name].append(node)
        for nodes in seen.values():
            if len(nodes) > 1:
                rv.append(nodes)

    return rv


async def upload_file_from_local(
    drive: Drive,
    path: Path,
    parent: Node,
    *,
    mime_type: str | None = None,
    media_info: MediaInfo | None = None,
    timeout: float | None = None,
) -> Node | None:
    # sanity check
    path = path.resolve()
    if not path.is_file():
        raise ValueError("invalid file")

    file_name = path.name
    total_file_size = path.stat().st_size
    if not mime_type:
        mime_type = _DEFAULT_FILE_MIME_TYPE

    async with drive.upload_file(
        name=file_name,
        parent=parent,
        size=total_file_size,
        mime_type=mime_type,
        media_info=media_info,
    ) as fout:
        with open(path, "rb") as fin:
            await _upload_retry(fin, fout, timeout)

    node = await fout.node()
    return node


async def _upload_retry(
    fin: BinaryIO, fout: WritableFile, timeout: float | None
) -> None:
    while True:
        try:
            await _upload_feed(fin, fout, timeout)
            break
        except TimeoutError:
            getLogger(__name__).exception("upload timeout, retry")

        await _upload_continue(fin, fout)


async def _upload_feed(
    fin: BinaryIO, fout: WritableFile, timeout: float | None
) -> None:
    while True:
        # Just in case the FS is unstable.
        async with asyncio.timeout(timeout):
            chunk = fin.read(_CHUNK_SIZE)
        if not chunk:
            break
        async with asyncio.timeout(timeout):
            await fout.write(chunk)
    async with asyncio.timeout(timeout):
        await fout.flush()


async def _upload_continue(fin: BinaryIO, fout: WritableFile) -> None:
    offset = await fout.tell()
    await fout.seek(offset)
    fin.seek(offset, os.SEEK_SET)


async def download_file_to_local(
    drive: Drive,
    node: Node,
    path: Path,
    *,
    timeout: float | None = None,
) -> Path:
    if node.is_directory:
        raise ValueError(f"cannot download a directory")

    if not path.is_dir():
        raise ValueError(f"{path} is not a directory")

    # check if exists
    complete_path = path.joinpath(node.name)
    if complete_path.is_file():
        return complete_path

    # exists but not a file
    if complete_path.exists():
        raise IsADirectoryError(complete_path)

    # if the file is empty, no need to download
    if node.size <= 0:
        open(complete_path, "w").close()
        return complete_path

    # resume download
    tmp_path = complete_path.parent.joinpath(f"{complete_path.name}.__tmp__")
    if tmp_path.is_file():
        offset = tmp_path.stat().st_size
        if offset > node.size:
            raise RuntimeError(
                f"local file size of `{complete_path}` is greater then remote"
                f" ({offset} > {node.size})"
            )
    elif tmp_path.exists():
        raise IsADirectoryError(complete_path)
    else:
        offset = 0

    if offset < node.size:
        async with drive.download_file(node) as fin:
            await fin.seek(offset)
            with open(tmp_path, "ab") as fout:
                await _download_retry(fin, fout, timeout)

    # rename it back if completed
    tmp_path.rename(complete_path)

    return complete_path


async def _download_retry(
    fin: ReadableFile, fout: BinaryIO, timeout: float | None
) -> None:
    while True:
        try:
            await _download_feed(fin, fout, timeout)
            break
        except TimeoutError:
            getLogger(__name__).exception("download timeout, retry")

        await _download_continue(fin, fout)


async def _download_feed(
    fin: ReadableFile, fout: BinaryIO, timeout: float | None
) -> None:
    while True:
        async with asyncio.timeout(timeout):
            chunk = await fin.read(_CHUNK_SIZE)
        if not chunk:
            break
        # Just in case the FS is unstable.
        async with asyncio.timeout(timeout):
            fout.write(chunk)
    # Just in case the FS is unstable.
    async with asyncio.timeout(timeout):
        fout.flush()


async def _download_continue(fin: ReadableFile, fout: BinaryIO) -> None:
    fout.flush()
    offset = fout.tell()
    await fin.seek(offset)



from abc import ABCMeta, abstractmethod
from collections.abc import (
    AsyncIterator,
    AsyncIterable,
    Awaitable,
    Callable,
)
from contextlib import AbstractAsyncContextManager
from dataclasses import dataclass
from datetime import datetime
from pathlib import PurePath
from typing import Literal, Self


__all__ = (
    "ChangeAction",
    "CreateFileService",
    "CreateFileServiceMiddleware",
    "CreateSnapshotService",
    "CreateSnapshotServiceMiddleware",
    "FileService",
    "Hasher",
    "MediaInfo",
    "Node",
    "PrivateDict",
    "ReadableFile",
    "RemoveAction",
    "UpdateAction",
    "WritableFile",
)


type PrivateDict = dict[str, str]


@dataclass(frozen=True, kw_only=True)
class Node:
    id: str
    parent_id: str | None
    name: str
    is_directory: bool
    is_trashed: bool
    ctime: datetime
    mtime: datetime
    mime_type: str
    hash: str
    size: int
    is_image: bool
    is_video: bool
    width: int
    height: int
    ms_duration: int
    private: PrivateDict | None


type RemoveAction = tuple[Literal[True], str]
type UpdateAction = tuple[Literal[False], Node]
type ChangeAction = RemoveAction | UpdateAction


class MediaInfo:
    @classmethod
    def image(cls, width: int, height: int) -> Self:
        return MediaInfo(is_image=True, width=width, height=height)

    @classmethod
    def video(cls, width: int, height: int, ms_duration: int) -> Self:
        return MediaInfo(
            is_video=True,
            width=width,
            height=height,
            ms_duration=ms_duration,
        )

    def __init__(
        self,
        *,
        is_image: bool = False,
        is_video: bool = False,
        width: int = 0,
        height: int = 0,
        ms_duration: int = 0,
    ) -> None:
        self._is_image = is_image
        self._is_video = is_video
        self._width = width
        self._height = height
        self._ms_duration = ms_duration

    def __str__(self) -> str:
        if self.is_image:
            return f"MediaInfo(is_image=True, width={self.width}, height={self.height})"
        if self.is_video:
            return f"MediaInfo(is_video=True, width={self.width}, height={self.height}, ms_duration={self.ms_duration})"
        return "MediaInfo()"

    @property
    def is_image(self) -> bool:
        return self._is_image

    @property
    def is_video(self) -> bool:
        return self._is_video

    @property
    def width(self) -> int:
        return self._width

    @property
    def height(self) -> int:
        return self._height

    @property
    def ms_duration(self) -> int:
        return self._ms_duration


class ReadableFile(AsyncIterable[bytes], metaclass=ABCMeta):
    """
    An async readable file interface.
    """

    @abstractmethod
    async def read(self, length: int) -> bytes:
        """
        Read at most `length` bytes.
        """

    @abstractmethod
    async def seek(self, offset: int) -> int:
        """
        Seek to `offset` position. Always starts from the begining.
        """

    @abstractmethod
    async def node(self) -> Node:
        """
        Get the node being read.
        """


class Hasher(metaclass=ABCMeta):
    """
    Hash calculator.
    """

    @abstractmethod
    async def update(self, data: bytes) -> None:
        """
        Put `data` into the stream.
        """

    @abstractmethod
    async def digest(self) -> bytes:
        """
        Get raw digest.
        """

    @abstractmethod
    async def hexdigest(self) -> str:
        """
        Get hex digest.
        """

    @abstractmethod
    async def copy(self) -> Self:
        """
        Return a copy to self.
        """


type CreateHasher = Callable[[], Awaitable[Hasher]]


class WritableFile(metaclass=ABCMeta):
    """
    An async writable file interface.
    """

    @abstractmethod
    async def tell(self) -> int:
        """
        Get current position.
        """

    @abstractmethod
    async def seek(self, offset: int) -> int:
        """
        Seek to `offset` position. Always starts from the begining.
        """

    @abstractmethod
    async def write(self, chunk: bytes) -> int:
        """
        Writes `chunk` to the stream.
        Returns bytes writen to buffer.
        """

    @abstractmethod
    async def flush(self) -> None:
        """
        Flush buffer.
        """

    @abstractmethod
    async def node(self) -> Node | None:
        """
        Get the wrote node. May be `None` if write failed.
        """


class Service(metaclass=ABCMeta):
    @property
    @abstractmethod
    def api_version(self) -> int:
        """
        Get competible API version for this class.
        """


type CreateService[T: Service] = Callable[[], AbstractAsyncContextManager[T]]
type CreateServiceMiddleware[T: Service] = Callable[[T], AbstractAsyncContextManager[T]]


class FileService(Service, metaclass=ABCMeta):
    """
    Provides actions to cloud drives.
    """

    @abstractmethod
    async def get_initial_cursor(self) -> str:
        """
        Get the initial check point.
        """

    @abstractmethod
    async def get_root(self) -> Node:
        """
        Fetch the root node.
        """

    @abstractmethod
    def get_changes(
        self,
        cursor: str,
    ) -> AsyncIterator[tuple[list[ChangeAction], str]]:
        """
        Fetch changes starts from `cursor`.

        Will be used like this:
        ```
        async for changes, next_cursor in self.fetch_changes('...'):
            ...
        ```
        So you should yield a page for every iteration.
        """

    @abstractmethod
    async def move(
        self,
        node: Node,
        *,
        new_parent: Node | None,
        new_name: str | None,
        trashed: bool | None,
    ) -> Node:
        """
        Changes the state of the given node:
        1. Rename the node.
        2. Move the node to another directory.
        3. Put the node to trash or restore it from trash.

        `node` is the node to be modified.

        `new_parent` is the new parent directory. `None` means don't move the node.

        `new_name` is the new node name. `None` means don't rename the node.

        `trashed` sets to `True` to put the node to trash. Sets to `False` will
        restore it. `None` means don't change trash state.
        """

    @abstractmethod
    async def purge_trash(self) -> None:
        """
        Purge everything in the trash.

        Should raise exception if failed.
        """

    @abstractmethod
    async def create_directory(
        self,
        name: str,
        parent: Node,
        *,
        exist_ok: bool,
        private: PrivateDict | None,
    ) -> Node:
        """
        Create a directory.

        `name` will be the name of the directory.

        `parent` should be a directory you want to put this directory in.

        If `exist_ok` is `False`, you should not create the directory if it is
        already exists, and raise an exception.

        `private` is an optional metadata, you can decide how to place this for
        each services.

        Will return the created node.
        """

    @abstractmethod
    def download_file(self, node: Node) -> AbstractAsyncContextManager[ReadableFile]:
        """
        Download the node.

        Will return a `ReadableFile` which is a file-like object.
        """

    @abstractmethod
    def upload_file(
        self,
        name: str,
        parent: Node,
        *,
        size: int | None,
        mime_type: str | None,
        media_info: MediaInfo | None,
        private: PrivateDict | None,
    ) -> AbstractAsyncContextManager[WritableFile]:
        """
        Upload a file.

        `name` is the name of the uploaded file.

        `parent` is the parent directory where this file upload to.

        `size` can be `None`, for cases that the file size is unavailable.
        e.g. The uploading file is from a stream.

        `mime_type`, `media_info` and `private` are optional. It is your choice
        to decide how to place these properties.
        """

    @abstractmethod
    async def get_hasher_factory(self) -> CreateHasher:
        """
        Get a hash calculator factory.
        """

    @abstractmethod
    async def is_authorized(self) -> bool:
        """
        Is OAuth 2.0 authorized.
        """

    @abstractmethod
    async def get_oauth_url(self) -> str:
        """
        Get OAuth 2.0 URL.
        """

    @abstractmethod
    async def set_oauth_token(self, token: str) -> None:
        """
        Set OAuth 2.0 token.
        """


type CreateFileService = CreateService[FileService]
type CreateFileServiceMiddleware = CreateServiceMiddleware[FileService]


class SnapshotService(Service, metaclass=ABCMeta):
    """
    Provides actions to cache file metadata.
    """

    @abstractmethod
    async def get_root(self) -> Node:
        """
        Get root directory as Node.
        """

    @abstractmethod
    async def set_root(self, node: Node) -> None:
        """
        Set root directory to the snapshot.
        """

    @abstractmethod
    async def get_current_cursor(self) -> str:
        """
        Get the current cursor. If no cursor present (e.g. the first run),
        should return an empty string.
        """

    @abstractmethod
    async def get_node_by_id(self, node_id: str) -> Node:
        """
        Get node by ID.
        """

    @abstractmethod
    async def get_node_by_path(self, path: PurePath) -> Node:
        """
        Resolve node by file path.
        """

    @abstractmethod
    async def resolve_path_by_id(self, node_id: str) -> PurePath:
        """
        Resolve absolute path by ID.
        """

    @abstractmethod
    async def get_child_by_name(self, name: str, parent_id: str) -> Node:
        """
        Get a child under the given parent by name.
        """

    @abstractmethod
    async def get_children_by_id(self, parent_id: str) -> list[Node]:
        """
        Get first-level children under a node.
        """

    @abstractmethod
    async def get_trashed_nodes(self) -> list[Node]:
        """
        Get trashed nodes.
        """

    @abstractmethod
    async def apply_changes(
        self,
        changes: list[ChangeAction],
        cursor: str,
    ) -> None:
        """
        Apply the given changes to the snapshot.
        """

    @abstractmethod
    async def find_nodes_by_regex(self, pattern: str) -> list[Node]:
        """
        Find node by regex.
        """


type CreateSnapshotService = CreateService[SnapshotService]
type CreateSnapshotServiceMiddleware = CreateServiceMiddleware[SnapshotService]


class Drive(metaclass=ABCMeta):
    """
    Interact with the drive.
    """

    @abstractmethod
    async def get_root(self) -> Node:
        """Get the root node."""

    @abstractmethod
    async def get_node_by_id(self, node_id: str) -> Node:
        """Get node by node id."""

    @abstractmethod
    async def get_node_by_path(self, path: PurePath) -> Node:
        """Get node by absolute path."""

    @abstractmethod
    async def resolve_path(self, node: Node) -> PurePath:
        """Resolve absolute path of the node."""

    @abstractmethod
    async def get_child_by_name(
        self,
        name: str,
        parent: Node,
    ) -> Node:
        """Get node by given name and parent."""

    @abstractmethod
    async def get_children(self, parent: Node) -> list[Node]:
        """Get the child node list of given node."""

    @abstractmethod
    async def get_trashed_nodes(self, flatten: bool = False) -> list[Node]:
        """Get trashed node list."""

    @abstractmethod
    async def find_nodes_by_regex(self, pattern: str) -> list[Node]:
        """Find nodes by name."""

    @abstractmethod
    def walk(
        self,
        node: Node,
        *,
        include_trashed: bool = False,
    ) -> AsyncIterator[tuple[Node, list[Node], list[Node]]]:
        """Traverse nodes from given node."""

    @abstractmethod
    async def create_directory(
        self,
        name: str,
        parent: Node,
        *,
        exist_ok: bool = False,
    ) -> Node:
        """Create a directory."""

    @abstractmethod
    def download_file(self, node: Node) -> AbstractAsyncContextManager[ReadableFile]:
        """Download file."""

    @abstractmethod
    def upload_file(
        self,
        name: str,
        parent: Node,
        *,
        size: int | None = None,
        mime_type: str | None = None,
        media_info: MediaInfo | None = None,
    ) -> AbstractAsyncContextManager[WritableFile]:
        """Upload file."""

    @abstractmethod
    async def purge_trash(self) -> None:
        """Purge everything in trash."""

    @abstractmethod
    async def move(
        self,
        node: Node,
        *,
        new_parent: Node | None = None,
        new_name: str | None = None,
        trashed: bool | None = None,
    ) -> Node:
        """Move or rename or trash the node."""

    @abstractmethod
    def sync(self) -> AsyncIterator[ChangeAction]:
        """Synchronize the snapshot.

        This is the ONLY function which will modify the snapshot.
        """

    @abstractmethod
    async def get_hasher_factory(self) -> CreateHasher:
        """Get a Hasher instance for checksum calculation."""

    @abstractmethod
    async def is_authorized(self) -> bool:
        """Is the drive authorized."""

    @abstractmethod
    async def get_oauth_url(self) -> str:
        """Get OAuth 2.0 URL"""

    @abstractmethod
    async def set_oauth_token(self, token: str) -> None:
        """Set OAuth 2.0 token"""



from pathlib import Path

from .types import Node


__all__ = (
    "DriveError",
    "InvalidServiceError",
    "NodeExistsError",
    "NodeNotFoundError",
    "IsADirectoryError",
    "UnauthorizedError",
)


class DriveError(Exception):
    pass


class InvalidServiceError(DriveError):
    pass


class NodeExistsError(DriveError):
    def __init__(self, node: Node) -> None:
        super().__init__(f"node already exists: {node.name}")
        self.node = node


class NodeNotFoundError(DriveError):
    def __init__(self, id: str) -> None:
        super().__init__(f"node not found: {id}")
        self.id = id


class IsADirectoryError(DriveError):
    def __init__(self, path: Path) -> None:
        super().__init__(f"{path} is a directory")
        self.path = path


class UnauthorizedError(DriveError):
    pass



from collections.abc import Awaitable
from pathlib import Path, PurePath

from .exceptions import NodeNotFoundError
from .types import Node


def resolve_path(
    from_: PurePath,
    to: PurePath,
) -> PurePath:
    rv = from_
    for part in to.parts:
        if part == ".":
            continue
        elif part == "..":
            rv = rv.parent
        else:
            rv = rv / part
    return rv


def normalize_path(path: PurePath) -> PurePath:
    if not path.is_absolute():
        raise ValueError("only accepts absolute path")
    rv: list[str] = []
    for part in path.parts:
        if part == ".":
            continue
        elif part == ".." and rv[-1] != "/":
            rv.pop()
        else:
            rv.append(part)
    return PurePath(*rv)


def is_valid_name(name: str) -> bool:
    if name.find("\\") >= 0:
        return False
    path = Path(name)
    return path.name == name


async def else_none(aw: Awaitable[Node]) -> Node | None:
    try:
        return await aw
    except NodeNotFoundError:
        return None
